﻿from typing import Optional, Tuple
import os
import re
import enum
import subprocess

from PyQt5 import QtCore
from PyQt5.QtCore import QObject, QTimer


__all__ = ["ExternalMediaState", "ScreenshotFileManager"]


class ExternalMediaState(enum.IntEnum):
    DISCONNECTED = 0
    CONNECTED = 1


class ScreenshotFileManager(QObject):
    external_media_state_changed = QtCore.pyqtSignal(ExternalMediaState)

    def __init__(
            self,
            directory_name: str,
            external_media_path: Optional[str],
            fallback_path: Optional[str],
            show_save_as_dialog: bool,
            external_media_check_frequency: float = 5.0,
            parent=None
    ) -> None:
        super().__init__(parent=parent)

        self.__directory_name = directory_name
        self.__fallback_path = fallback_path

        self.__external_media_path = external_media_path

        self.__external_media_name = None
        self.__screenshot_number = None

        self.__show_save_as_dialog = show_save_as_dialog

        self.__external_media_watchdog = None

        self.__external_media_watchdog = _ExternalMediaWatchdog(
            check_frequency=external_media_check_frequency,
            external_media_path=self.__external_media_path,
            directory_name=self.__directory_name
        )
        self.__external_media_watchdog \
            .external_media_state_changed \
            .connect(self.__on_external_media_state_changed)

    @property
    def external_media_state(self) -> ExternalMediaState:
        return ExternalMediaState.DISCONNECTED \
            if self.__external_media_name is None \
            else ExternalMediaState.CONNECTED

    @property
    def lightweight_workers(self) -> list:
        workers = []
        if self.__external_media_watchdog is not None:
            workers.append(self.__external_media_watchdog)
        return workers

    def get_reference_path(self):
        self.__check_external_media()
        if self.__screenshot_number is None:
            return
        if self.__external_media_name is None:
            if self.__fallback_path is None:
                return
            reference_path = os.path.join(
                os.path.expanduser(self.__fallback_path),
                "Reference")
        else:
            reference_path = os.path.join(
                os.path.expanduser(self.__external_media_path),
                self.__external_media_name,
                "Reference")
        return reference_path

    def get_next_screenshot_file_name(self) -> Optional[str]:
        self.__check_external_media()
        if self.__screenshot_number is None:
            return

        if self.__external_media_name is None:
            screenshot_path = os.path.join(
                os.path.expanduser(self.__fallback_path),
                self.__directory_name)
        else:
            screenshot_path = os.path.join(
                os.path.expanduser(self.__external_media_path),
                self.__external_media_name,
                self.__directory_name)
        return os.path.join(
            screenshot_path,
            "{:0>6d}.png".format(self.__screenshot_number)
        )

    def commit_screenshot(self) -> None:
        self.__screenshot_number += 1
        try:
            os.sync()
        except Exception:
            pass

    def __check_external_media(self) -> None:
        media_was_connected = self.__external_media_name is not None

        self.__external_media_name, self.__screenshot_number =\
            _get_next_screenshot_dst(
                external_media_name=self.__external_media_name,
                external_media_path=self.__external_media_path,

                directory_name=self.__directory_name,
                fallback_path=self.__fallback_path,

                calculate_screenshot_number=True,
                screenshot_number=self.__screenshot_number
            )

        media_is_connected = self.__external_media_name is not None

        if media_is_connected != media_was_connected:
            external_media_state = ExternalMediaState.CONNECTED \
                if media_is_connected else ExternalMediaState.DISCONNECTED
            self.external_media_state_changed.emit(external_media_state)

    @QtCore.pyqtSlot(ExternalMediaState, int)
    def __on_external_media_state_changed(
            self,
            state: ExternalMediaState, screenshot_number: int) -> None:
        self.external_media_state_changed.emit(state)
        self.__screenshot_number = screenshot_number


class _ExternalMediaWatchdog(QObject):
    # (state, next screenshot number)
    external_media_state_changed = QtCore.pyqtSignal(ExternalMediaState, int)

    def __init__(
            self,

            check_frequency: Optional[float],

            external_media_path: Optional[str],
            directory_name: str,

            external_media_name: Optional[str] = None,
            screenshot_number: Optional[int] = None,

            parent=None
    ) -> None:
        super().__init__(parent=parent)

        self.__external_media_name = external_media_name
        self.__external_media_path = external_media_path
        self.__directory_name = directory_name
        self.__screenshot_number = screenshot_number

        self.__timer = QTimer(parent=self)
        self.__timer.timeout.connect(self.__on_timeout)
        # in milliseconds
        self.__timer.setInterval(int(1000.0 / check_frequency))
        self.__timer.setSingleShot(False)

    @QtCore.pyqtSlot()
    def start(self) -> None:
        self.__timer.start()

    @QtCore.pyqtSlot()
    def stop(self) -> None:
        self.__timer.stop()

    @QtCore.pyqtSlot()
    def __on_timeout(self) -> None:

        media_was_connected = self.__external_media_name is not None

        self.__external_media_name, self.__screenshot_number =\
            _get_next_screenshot_dst(
                external_media_name=self.__external_media_name,
                external_media_path=self.__external_media_path,

                directory_name=self.__directory_name,
                calculate_screenshot_number=True,
                screenshot_number=self.__screenshot_number
            )

        media_is_connected = self.__external_media_name is not None

        if media_is_connected != media_was_connected:
            external_media_state = \
                ExternalMediaState.CONNECTED \
                if media_is_connected else \
                ExternalMediaState.DISCONNECTED
            self.external_media_state_changed.emit(
                external_media_state, self.__screenshot_number)


def _get_next_screenshot_dst(
        external_media_name: Optional[str],
        external_media_path: Optional[str],

        directory_name: Optional[str] = None,
        fallback_path: Optional[str] = None,

        calculate_screenshot_number: bool = False,
        screenshot_number: Optional[int] = None
) -> Tuple[Optional[str], Optional[int]]:
    """Returns new external media name.
     If external media found and next screenshot number."""

    if external_media_path is None:
        # External media not supported
        external_media_name = None
    else:
        # Try to search new external media or fallback to internal
        screenshot_number = None
        external_media_name = __try_find_new_external_media(external_media_path)
        if external_media_name is None:
            return None, None

    if not is_mounted(os.path.join(external_media_path, external_media_name)):
        return None, None

    if screenshot_number is None and directory_name is not None:
        screenshot_directory_path = None

        if external_media_name is not None:
            try:
                screenshot_directory_path = __prepare_screenshot_directory(
                    os.path.join(external_media_path,
                                 external_media_name, directory_name)
                )
            except Exception:
                pass

            if screenshot_directory_path is None:
                return None, None

        if screenshot_directory_path is None:
            calculate_screenshot_number = False

        if calculate_screenshot_number:
            screenshot_number = __get_next_screenshot_file_number(
                screenshot_directory_path)

    return external_media_name, screenshot_number


def is_mounted(mount_path) -> bool:
        import sys
        import platform
        if sys.platform != "linux":
            return True

        if platform.processor() == "x86_64":
            return True

        process = subprocess.Popen(["mount"], stdout=subprocess.PIPE)
        for line in iter(process.stdout.readline, ""):
            chunks = line.decode("utf8").split(" ")
            if (len(chunks) < 3):
                return False
            if chunks[2] == mount_path:
                for a in os.listdir("/dev"):
                    if os.path.join("/dev", a) == chunks[0]:
                        return True
        return False


def __try_find_new_external_media(external_media_path: str) -> Optional[str]:
    try:
        dirs = os.listdir(external_media_path)
        if len(dirs) == 0:
            return None

        # Note: directories are not ordered
        for external_media_name in dirs:
            return external_media_name
    except Exception as e:
        print("When searching external media caught: {}.".format(e))


def __prepare_screenshot_directory(path: Optional[str]) -> str:
    directory = os.path.expanduser(path)
    try:
        os.makedirs(directory, exist_ok=True)
    except Exception as e:
        return None
    return directory


__screenshot_file_name_pattern = re.compile("^(\\d+)\\.png$")


def __get_next_screenshot_file_number(directory_path: str) -> int:
    max_screenshot_number = 0
    for screenshot_file_path in os.listdir(directory_path):
        screenshot_file_name = os.path.basename(screenshot_file_path)

        m = __screenshot_file_name_pattern.match(screenshot_file_name)
        if m is not None:
            max_screenshot_number = max(max_screenshot_number, int(m.group(1)))

    return max_screenshot_number + 1
